/*!
 * (c) 2006-2007 EPFL, Lausanne, Switzerland
 * Thomas Lochmatter
 */

#include <fcntl.h>
#include <linux/i2c-dev.h>

// Variables filled in by command line parser
int i2c_fd;
#define MESSAGES_LEN 128
struct i2c_msg messages[MESSAGES_LEN];
int messages_nw = 0;
#define WRITE_BUFFER_LEN 1024
char write_buffer[WRITE_BUFFER_LEN];
int write_buffer_nw = 0;
int write_buffer_start = 0;
#define READ_BUFFER_LEN 1024
char read_buffer[READ_BUFFER_LEN];
int read_buffer_nw = 0;
int repeat_count = 1;
int dev = -1;
int verbose = 0;
int quiet = 0;

/*--------------------------------------------------------------------*/
/*! Prints the help text.
 */
void help() {
	printf("Usage: i2c [-v|--verbose] [-q|--quiet] [-r|--repeat COUNT] [d|device] DEVICE [w|write BYTE1 BYTE2 ...] [r|read LEN]\n");
	printf("\n");
	printf("Examples:\n");
	printf("  Writing 155 to register 5 of device 0x50 (if this device is register-based):\n");
	printf("    i2c 0x50 write 5 155\n");
	printf("  Reading register 5 of device 0x50 (if this device is register-based):\n");
	printf("    i2c 0x50 write 5 read 1\n");
	printf("  Sending a packet of 8 bytes to device 50:\n");
	printf("    i2c 0x50 write 51 23 166 255 65 0 34 66\n");
	printf("  Reading 8 bytes from device 50:\n");
	printf("    i2c 0x50 read 8\n");
	printf("  The same, but 10 times:\n");
	printf("    i2c -r 10 0x50 read 8\n");
	printf("  The same, but without any output:\n");
	printf("    i2c -q -r 10 0x50 read 8\n");
	printf("  Reading 2 bytes from device 0x40, writing to 0x50, then reading again from 0x40:\n");
	printf("    i2c device 0x40 read 2 device 0x50 write 51 23 device 0x40 read 2\n");
	printf("  To see what is going on, add a -v as first parameter.\n");
}

int add_read_message(int len) {
	messages[messages_nw].addr  = dev;
	messages[messages_nw].flags = I2C_M_RD;
	messages[messages_nw].buf   = &(read_buffer[read_buffer_nw]);
	messages[messages_nw].len   = len;

	if (verbose) {
		printf("Read message from %d (0x%x): start at %d, length %d bytes\n", dev, dev, read_buffer_nw, len);
	}

	messages_nw++;
	if (messages_nw > MESSAGES_LEN) {
		printf("Max number of messages is is %d.");
		exit(1);
	}

	read_buffer_nw += len;
	if (read_buffer_nw > READ_BUFFER_LEN) {
		printf("Max total read buffer is %d bytes.", READ_BUFFER_LEN);
		exit(1);
	}
}

int add_write_message() {
	int i;

	if (write_buffer_nw - write_buffer_start == 0) {
		return;
	}
	messages[messages_nw].addr  = dev;
	messages[messages_nw].flags = 0;
	messages[messages_nw].buf   = &(write_buffer[write_buffer_start]);
	messages[messages_nw].len   = write_buffer_nw - write_buffer_start;
	write_buffer_start = write_buffer_nw;

	if (verbose) {
		printf("Write message to %d (0x%x):", dev, dev);
		for (i = write_buffer_start; i < write_buffer_nw; i++) {
			printf(" %d", write_buffer[i]);
		}
		printf("\n");
	}

	messages_nw++;
	if (messages_nw > MESSAGES_LEN) {
		printf("Max number of messages is is %d.");
		exit(1);
	}
}

int add_write_byte(int val) {
	write_buffer[write_buffer_nw++] = val;

	if (write_buffer_nw > WRITE_BUFFER_LEN) {
		printf("Max total write buffer is %d bytes.", WRITE_BUFFER_LEN);
		exit(1);
	}
}

/*--------------------------------------------------------------------*/
/*! Main program.
 */
int run(int argc, char *argv[]) {
	int i, k;
	int repeat;
#define STATE_OPERATION 0
#define STATE_REPEAT 1
	int state = STATE_OPERATION;
#define OPERATION_DEVICE 0
#define OPERATION_READ 1
#define OPERATION_WRITE 2
	int operation = OPERATION_DEVICE;
	struct i2c_rdwr_ioctl_data msgset;

	i2c_fd = open("/dev/i2c/0", O_RDWR);
	if (i2c_fd < 0) {
		printf("Unable to open I2C bus (/dev/i2c/0): res=%d", i2c_fd);
		exit(1);
	}

	// Initialize variables
	write_buffer_nw = 0;
	write_buffer_start = 0;
	read_buffer_nw = 0;
	messages_nw = 0;
	repeat_count = 1;
	dev = -1;
	verbose = 0;
	quiet = 0;

	// Read command line
	for (i = 1; i < argc; i++) {
		if ((strcmp(argv[i], "-r") == 0) || (strcmp(argv[i], "--repeat") == 0)) {
			state = STATE_REPEAT;
		} else if ((strcmp(argv[i], "-v") == 0) || (strcmp(argv[i], "--verbose") == 0)) {
			verbose = 1;
		} else if ((strcmp(argv[i], "-q") == 0) || (strcmp(argv[i], "--quiet") == 0)) {
			quiet = 1;
		} else if ((strcmp(argv[i], "w") == 0) || (strcmp(argv[i], "write") == 0)) {
			add_write_message();
			operation = OPERATION_WRITE;
		} else if ((strcmp(argv[i], "r") == 0) || (strcmp(argv[i], "read") == 0)) {
			add_write_message();
			operation = OPERATION_READ;
		} else if ((strcmp(argv[i], "d") == 0) || (strcmp(argv[i], "device") == 0)) {
			add_write_message();
			operation = 0;
		} else if (state == STATE_OPERATION) {
			if (operation == OPERATION_DEVICE) {
				dev = atoihex(argv[i]);
			} else if (operation == OPERATION_READ) {
				add_read_message(atoihex(argv[i]));
			} else if (operation == OPERATION_WRITE) {
				add_write_byte(atoihex(argv[i]));
			}
		} else if (state == STATE_REPEAT) {
			repeat_count = atoihex(argv[i]);
			state = STATE_OPERATION;
		}
	}

	add_write_message();

	if ((state != STATE_OPERATION) || (dev < 0) || (messages_nw < 1)) {
		help();
		exit(1);
	}

	// Prepare message set
	msgset.msgs = messages;
	msgset.nmsgs = messages_nw;
	if (verbose) {
		printf("\n");
		printf("Number of messages: %d\n", msgset.nmsgs);
		printf("Number of repetitions: %d\n", repeat_count);
		printf("\n");
	}

	// Send/receive data
	for (repeat = 0; repeat < repeat_count; repeat++) {
		// Call
		int res = ioctl(i2c_fd, I2C_RDWR, &msgset);

		// Print results
		if (verbose) {
			printf("Trial %d: res=%d\n", repeat + 1, res);
			for (i = 0; i < messages_nw; i++) {
				if (messages[i].flags == I2C_M_RD) {
					printf("  Read from %d (0x%x):\n", messages[i].addr);
					for (k = 0; k < messages[i].len; k++) {
						printf("%d ", messages[i].buf[k]);
					}
					printf("\n");
				} else {
					printf("  Written to %d (0x%x):\n", messages[i].addr);
					for (k = 0; k < messages[i].len; k++) {
						printf("%d ", messages[i].buf[k]);
					}
					printf("\n");
				}
			}
		} else if (quiet) {
		} else {
			if (read_buffer_nw > 0) {
				for (i = 0; i < read_buffer_nw; i++) {
					printf("%d ", read_buffer[i]);
				}
				printf("\n");
			}
		}
	}

	return 0;
}

int atoihex(char *arg) {
	int n;

	n = atoi(arg);
	sscanf(arg, "0x%x", &n);
	return n;
}

/*--------------------------------------------------------------------*/
/*! Initialize and run the program.
 */
int main(int argc, char *argv[]) {
	// Help
	if ((argc == 2) && ((strcmp(argv[1], "help") == 0) || (strcmp(argv[1], "-help") == 0) || (strcmp(argv[1], "-h") == 0) || (strcmp(argv[1], "--help") == 0))) {
		help();
		return 0;
	}

	return run(argc, argv);
}

