/*!
 * (c) 2006-2007 EPFL, Lausanne, Switzerland
 * Thomas Lochmatter
 */

#include <stdio.h>
#include <fcntl.h>
#include <linux/i2c-dev.h>
#include "i2cal.h"

// Variables filled in by command line parser
int repeat_count = 1;
int dev = -1;
int reg = -1;
int verbose = 0;
int quiet = 0;
int sleep_usec = 0;
int reset = 1;

/*--------------------------------------------------------------------*/
/*! Prints the help text.
 */
void help() {
	printf("Usage: i2c_readstream [-v|--verbose] [-q|--quiet] [-s|--sleep USEC] [--no-reset] DEVICE QUERYREGISTER [DATAREGISTER=QUERYREGISTER+1]\n");
	printf("\n");
	printf("Examples:\n");
	printf("  Streaming from register 0x86 (query) and 0x87 (data) from device 0x50:\n");
	printf("    i2c_readstream 0x50 0x86\n");
	printf("  The same, but with sleeps of 10ms between queries:\n");
	printf("    i2c_readstream 0x50 0x86 --sleep 10000\n");
	printf("  The same, but without resetting the stream at the beginning:\n");
	printf("    i2c_readstream 0x50 0x86 --sleep 10000 --no-reset\n");
	printf("  To see what is going on, add a -v as first parameter.\n");
}

int atoihex(char *arg) {
	int n;

	n = atoi(arg);
	sscanf(arg, "0x%x", &n);
	return n;
}

/*--------------------------------------------------------------------*/
/*! Main program.
 */
int run(int argc, char *argv[]) {
	int i, k;
	int repeat;
#define STATE_OPERATION 0
#define STATE_REPEAT 1
#define STATE_SLEEP 2
	int state = STATE_OPERATION;
#define OPERATION_DEVICE 0
#define OPERATION_REGISTER 1
	int operation = OPERATION_DEVICE;
	int streamstatus;
	struct i2c_msg *msg_streamstatus;
	struct i2c_msg *msg_data;
#define STREAM_RESET 0
#define STREAM_GETSTATUS 1
#define STREAM_GETDATA 2

	// Open the I2C bus
	if (i2cal_open() == 0) {
		printf("Unable to open I2C bus (/dev/i2c/0): res=%d", i2cal_fd);
		exit(1);
	}

	// Initialize variables
	repeat_count = 1;
	dev = -1;
	reg = -1;
	verbose = 0;
	quiet = 0;
	sleep_usec = 0;
	reset = 1;

	// Read command line
	for (i = 1; i < argc; i++) {
		if ((strcmp(argv[i], "-r") == 0) || (strcmp(argv[i], "--repeat") == 0)) {
			state = STATE_REPEAT;
		} else if ((strcmp(argv[i], "-v") == 0) || (strcmp(argv[i], "--verbose") == 0)) {
			verbose = 1;
		} else if ((strcmp(argv[i], "-q") == 0) || (strcmp(argv[i], "--quiet") == 0)) {
			quiet = 1;
		} else if ((strcmp(argv[i], "-s") == 0) || (strcmp(argv[i], "--sleep") == 0)) {
			state = STATE_SLEEP;
		} else if (strcmp(argv[i], "--no-reset") == 0) {
			reset = 0;
		} else if (state == STATE_OPERATION) {
			if (operation == OPERATION_DEVICE) {
				dev = atoihex(argv[i]);
				operation = OPERATION_REGISTER;
			} else if (operation == OPERATION_REGISTER) {
				reg = atoihex(argv[i]);
			}
		} else if (state == STATE_REPEAT) {
			repeat_count = atoihex(argv[i]);
			state = STATE_OPERATION;
		} else if (state == STATE_SLEEP) {
			sleep_usec = atoihex(argv[i]);
			state = STATE_OPERATION;
		}
	}

	if ((state != STATE_OPERATION) || (dev < 0) || (reg < 0)) {
		help();
		exit(1);
	}

	// Send the reset command
	if (reset) {
		if (verbose) {
			printf("Resetting stream ...\n");
		}
		i2cal_start();
		i2cal_writedata_uchar(reg);
		i2cal_writedata_uchar(STREAM_RESET);
		i2cal_write(dev);
		i2cal_commit();
	}

	// Read the stream
	if (verbose) {
		printf("Reading from stream ...\n");
	}
	while (1) {
		// Read the (guaranteed) available amount of data
		i2cal_start();
		i2cal_writedata_uchar(reg);
		i2cal_writedata_uchar(STREAM_GETSTATUS);
		i2cal_write(dev);
		msg_streamstatus = i2cal_read(dev, 1);
		i2cal_commit();

		// Quit on error (HSB of the message length indicates error status)
		streamstatus = msg_streamstatus->buf[0];
		if (streamstatus == 129) {
			printf("ERROR: Stream error %d: Buffer overrun!\n", streamstatus);
			break;
		} else if (streamstatus & 0x80) {
			printf("ERROR: Stream error %d.\n", streamstatus);
			break;
		} else {
			if (verbose) {
				printf("Available bytes: %d\n", streamstatus);
			}
		}

		// Read data from the stream
		i2cal_start();
		i2cal_writedata_uchar(reg);
		i2cal_writedata_uchar(STREAM_GETDATA);
		i2cal_write(dev);
		msg_data = i2cal_read(dev, streamstatus);
		i2cal_commit();

		// Print that data
		if (! quiet) {
			for (i = 0; i < streamstatus; i++) {
				printf("%d ", msg_data->buf[i]);
			}
			printf("\n");
			fflush(stdout);
		}

		// Sleep
		if (sleep_usec > 0) {
			usleep(sleep_usec);
		}
	}

	return 0;
}

/*--------------------------------------------------------------------*/
/*! Initialize and run the program.
 */
int main(int argc, char *argv[]) {
	// Help
	if ((argc == 2) && ((strcmp(argv[1], "help") == 0) || (strcmp(argv[1], "-help") == 0) || (strcmp(argv[1], "-h") == 0) || (strcmp(argv[1], "--help") == 0))) {
		help();
		return 0;
	}

	return run(argc, argv);
}
