/*!
 * (c) 2006-2007 EPFL, Lausanne, Switzerland
 * Thomas Lochmatter
 */

#include "khepera3.h"
#include "commandline.h"
#include "motion_arrowkeys.h"
#include "interactiveinput.h"
#include <termios.h>
#include <stdlib.h>

// Level of verbosity (0=quiet, 1=normal, 2=verbose)
int verbosity = 1;
// Loop speed
int wait_us = 50 * 1000;

// Prints the help text.
void help() {
	printf("Allows to control the robot with arrow keys.\r\n");
	printf("Usage: motion_arrowkeys [OPTIONS] [ALGORITHM]\r\n");
	printf("  -w --wait-us US    Sets the waiting time in the perception-to-action loop (default: 50000)\r\n");
	printf("  -q --quiet         Be quiet (no output)\r\n");
	printf("  -v --verbose       Be verbose\r\n");
	printf("\r\n");
	printf("Keys\r\n");
	printf("  +-         increase/decrease speed step\r\n");
	printf("  123456789  set speed and direction\r\n");
	printf("  qweasdyxc  set speed and direction\r\n");
	printf("  <enter>    quit\r\n");
}

int motion_arrowkeys_run(struct sMotionArrowKeys * motion_arrowkeys) {
	int key, timediff;
	struct termios tty;
	struct timeval timprev;
	struct timeval timnext;
	int speed_left, speed_right;

	// Put the wheels in normal (control) mode
	khepera3_drive_start();

	gettimeofday(&timprev, 0);
	while (1) {
		// Wait for the next key
		key = interactiveinput_waitkey(wait_us);

		// Quit on ENTER, SPACE or 0
		if ((key == 10) || (key == 32) || (key == '0')) {
			break;
		}

		// Update speed
		gettimeofday(&timnext, 0);
		timediff = (timnext.tv_sec - timprev.tv_sec) * 1000000 + (timnext.tv_usec - timprev.tv_usec);
		timprev = timnext;
		motion_arrowkeys_update(motion_arrowkeys, key, timediff);

		// Set motor speed
		speed_left = motion_arrowkeys_getspeedleft(motion_arrowkeys);
		speed_right = motion_arrowkeys_getspeedright(motion_arrowkeys);
		printf("$SPEED,%d,%d\r\n", speed_left, speed_right);
		if ((abs(speed_left) < 1) && (abs(speed_right) < 1)) {
			khepera3_drive_set_speed(0, 0);
		} else {
			khepera3_drive_set_speed(speed_left, speed_right);
		}
	}
	return 0;
}

// Main program.
int main(int argc, char *argv[]) {
	struct sMotionArrowKeys motion_arrowkeys;
	struct sCommandLineOption *option_speed;
	struct sCommandLineOption *option_wait_us;

	// Command line parsing
	commandline_init();
	option_wait_us = commandline_option_register("-w", "--wait-us", cCommandLine_Option_Value);
	commandline_parse(argc, argv);

	// Help
	if (commandline_option_provided("-h", "--help")) {
		help();
		exit(1);
	}

	// Initialization
	khepera3_init();
	interactiveinput_init();
	motion_arrowkeys_init(&motion_arrowkeys);

	// Read arguments
	if (option_wait_us->value) {
		wait_us = atoi(option_wait_us->value);
	}
	if (commandline_option_provided("-q", "--quiet")) {
		verbosity = 0;
	}
	if (commandline_option_provided("-v", "--verbose")) {
		verbosity = 2;
	}

	// Run the algorithm
	motion_arrowkeys_run(&motion_arrowkeys);
	return 0;
}

