/*!
 * (c) 2006-2007 EPFL, Lausanne, Switzerland
 * Thomas Lochmatter
 */

#include "khepera3.h"
#include "commandline.h"

// Algorithm parameters and results
struct sFollowLine {
	int speed;
	int aggressivity;
	long int wait_us;
};

// Level of verbosity (0=quiet, 1=normal, 2=verbose)
int verbosity = 1;

// Prints the help text.
void help() {
	printf("Follows a line on the ground with a simple Braitenberg algorithm.\r\n");
	printf("Usage: motion_followline [OPTIONS]\r\n");
	printf("  --speed SPEED        Sets the forward speed (default: 10000)\r\n");
	printf("  --aggressivity AGG   Sets the aggressivity (default: 4)\r\n");
	printf("  -w --wait-us US      Sets the waiting time in the perception-to-action loop (default: 50000)\r\n");
	printf("  -q --quiet           Be quiet (no output)\r\n");
	printf("  -v --verbose         Be verbose\r\n");
	printf("  Positive aggressivity will follow black/dark lines.\r\n");
	printf("  Negative aggressivity will follow white/bright lines.\r\n");
	printf("Output:\r\n");
	printf("  $IRFLOOR,sensor_left,sensor_right\r\n");
	printf("  $SPEED,speed_left,speed_right\r\n");
}

void followline_init(struct sFollowLine *followline) {
	followline->speed = 10000;
	followline->aggressivity = 4;
	followline->wait_us = 50*1000;
}

// Follows a line.
void followline_run(struct sFollowLine *followline) {
	int speeddiff = 0;

	// Put the wheels in normal (control) mode
	khepera3_drive_start();

	while (1) {
		// Get new IR proximity values
		khepera3_infrared_proximity();

		// Calculate the speed difference
		speeddiff = 0;
		speeddiff += khepera3.infrared_proximity.sensor[cKhepera3SensorsInfrared_FloorLeft] * followline->aggressivity;
		speeddiff -= khepera3.infrared_proximity.sensor[cKhepera3SensorsInfrared_FloorRight] * followline->aggressivity;

		// Set the speed
		khepera3_drive_set_speed(followline->speed - speeddiff, followline->speed + speeddiff);

		// Print values
		if (verbosity > 1) {
			printf("$IRFLOOR,%d,%d\r\n",
			       khepera3.infrared_proximity.sensor[cKhepera3SensorsInfrared_FloorLeft],
			       khepera3.infrared_proximity.sensor[cKhepera3SensorsInfrared_FloorRight]
			      );
		}
		if (verbosity > 0) {
			printf("$SPEED,%d,%d\r\n", followline->speed - speeddiff, followline->speed + speeddiff);
		}

		// Wait
		usleep(followline->wait_us);
	}
}

// Main program.
int main(int argc, char *argv[]) {
	struct sFollowLine followline;
	struct sCommandLineOption *option_speed;
	struct sCommandLineOption *option_aggressivity;
	struct sCommandLineOption *option_wait_us;

	option_speed = commandline_option_register("-s", "--speed", cCommandLine_Option_Value);
	option_aggressivity = commandline_option_register("-a", "--aggressivity", cCommandLine_Option_Value);
	option_wait_us = commandline_option_register("-w", "--wait-us", cCommandLine_Option_Value);

	// Command line parsing
	commandline_init();
	commandline_parse(argc, argv);

	// Help
	if (commandline_option_provided("-h", "--help")) {
		help();
		exit(1);
	}

	// Initialization
	khepera3_init();
	followline_init(&followline);

	// Read command line arguments
	if (option_speed->value) {
		followline.speed = atoi(option_speed->value);
	}
	if (option_aggressivity->value) {
		followline.aggressivity = atoi(option_aggressivity->value);
	}
	if (option_wait_us->value) {
		followline.wait_us = atoi(option_wait_us->value);
	}
	if (commandline_option_provided("-q", "--quiet")) {
		verbosity = 0;
	}
	if (commandline_option_provided("-v", "--verbose")) {
		verbosity = 2;
	}

	// Follow the line
	followline_run(&followline);
	return 0;
}
