/*!
 * (c) 2006-2007 EPFL, Lausanne, Switzerland
 * Loic Matthey, Thomas Lochmatter
 */

#include "khepera3.h"
#include "commandline.h"
#include <math.h>

// Algorithm parameters
struct sSpiral {
	int speed;
	long int wait_us;
	float a;
	float time;
	float time_increment;
};

// Level of verbosity (0=quiet, 1=normal, 2=verbose)
int verbosity = 1;


void help() {
	printf("Moves the robot along a archimedean spiral of increasing radius.\r\n");
	printf("Usage: spiralmotion [OPTIONS]\r\n");
	printf("  -a A                  Sets A, which defines the spiral gap (default: 5000)\r\n");
	printf("  -s --speed SPEED      Sets the speed (default: 10000)\r\n");
	printf("  -w --wait-us US       Waits US us in the main loop (default: 20000)\r\n");
	printf("  -q --quiet           Be quiet (no output)\r\n");
	printf("  -v --verbose         Be verbose\r\n");
	printf("Output:\r\n");
	printf("  $SPIRAL,TIME,SPEED_DIFF\r\n");
}

int spiral_init(struct sSpiral *spiral) {
	spiral->speed = 10000;
	spiral->wait_us = 20 * 1000;
	spiral->a = 5000;
	spiral->time = 1.0;
	spiral->time_increment = 0.1;
}

int spiral_run(struct sSpiral *spiral) {
	float speed_diff;
	int speed_left, speed_right;

	// Put the wheels in normal (control) mode
	khepera3_drive_start();

	while (1) {
		// This speed difference makes that the resulting trajectory is an archimedean spiral
		speed_diff = spiral->a / sqrtf(spiral->time);
		spiral->time += spiral->time_increment;

		// Calculate and set speeds
		speed_left = (int)floorf(spiral->speed - speed_diff + 0.5);
		speed_right = (int)floorf(spiral->speed + speed_diff + 0.5);
		khepera3_drive_set_speed(speed_left, speed_right);

		// Report
		if (verbosity > 1) {
			printf("$SPIRAL,%f,%f\r\n", time, speed_diff);
		}

		// Wait
		usleep(spiral->wait_us);
	}

	return 0;
}

// Main program.
int main(int argc, char *argv[]) {
	struct sCommandLineOption *option_wait_us;
	struct sCommandLineOption *option_speed;
	struct sCommandLineOption *option_a;
	struct sSpiral spiral;

	// Command line parsing
	commandline_init();
	option_wait_us = commandline_option_register("-w", "--wait-us", cCommandLine_Option_Value);
	option_speed = commandline_option_register("-s", "--speed", cCommandLine_Option_Value);
	option_a = commandline_option_register("-a", "--gap", cCommandLine_Option_Value);
	commandline_parse(argc, argv);

	// Help
	if (commandline_option_provided("-h", "--help")) {
		help();
		exit(1);
	}

	// Initialization
	khepera3_init();
	spiral_init(&spiral);

	// Read arguments
	if (option_wait_us->value) {
		spiral.wait_us = atoi(option_wait_us->value);
	}
	if (option_speed->value) {
		spiral.speed = atoi(option_speed->value);
	}
	if (option_a->value) {
		spiral.a = atoi(option_a->value);
	}
	if (commandline_option_provided("-q", "--quiet")) {
		verbosity = 0;
	}
	if (commandline_option_provided("-v", "--verbose")) {
		verbosity = 2;
	}

	// Run server
	spiral_run(&spiral);
	return 0;
}
